# Search for gravitational-lensing signatures in the full third observing run of the LIGO–Virgo network
## Python script for "Constraints on Merger Rate Models based on (non-)detection of lensing" | Figure 5

##########
#"Search for gravitational-lensing signatures in the full third observing run of the LIGO–Virgo network" is the second paper from the LIGO, 
# Virgo, and KAGRA (LVK) Collaborations searching for possible lensing features from gravitational waves during the full third observing run
# (O3a + O3b) which lasted from 1st April # 2019 to 27th March 2020.
##########

##############
# What is lensing of gravitational waves?
##############

# General relativity tells us that gravity curves space and time, affecting everything that carries mass or energy, and gravitational waves 
# are no exception. Therefore, we expect that as they travel cosmological distances, gravitational waves will inevitably encounter massive 
# astrophysical objects that would act as lenses, bending their trajectories and even distorting the shapes of their signals. In the paper, 
# we have implemented multiple analyses to search for the still undetected lensing signatures in the gravitational-wave signals detected in 
# O3a+b. 

# This code is part of the data release for the paper. It focuses on the analysis of constraints from non-detection of lensing on the merger rate density models, and shows you how to reproduce  the results shown in Figure 5 of the paper.

#######################
# \<Analysis Name\>
#######################

#############
# \<Brief introduction to the analysis\>
#############

import numpy as np
import pylab as plt

# Constants
h_H0=0.6727   ## Matched to Oguri et al. 2018
Gpc3            = 1.e9
R0h67           = 22./Gpc3 ## Mpc^3 / yr # Local value of merger rate
R0              = R0h67/h_H0**3 ### convert rate to h^3/Mpc^3/yr

##################################### 
# Source merger rate density model
##################################### 

RUnnormed = lambda z, kappa, gamma, zp: (1.+z)**kappa/(1.+((1.+z)/(1.+zp))**(gamma+kappa))
Rz        = lambda z, kappa, gamma, zp, R0: R0 * RUnnormed(z,kappa,gamma,zp) / RUnnormed(0,kappa,gamma,zp)

# PowerLaw only for  https://ldas-jobs.ligo.caltech.edu/~daniel.wysocki/o3b-population-data/analyses/PowerLawPeak/o1o2o3_mass_c_iid_mag_iid_tilt_powerlaw_redshift_result.json
RUnnPL = lambda z, kappa: (1.+z)**kappa
RzPL   = lambda z, kappa, R0: R0 * RUnnPL(z,kappa) / RUnnPL(0,kappa)


##################################### 
# - [ ] Computing 90% confidence interval for given weights
##################################### 
def computeLimits(x, W):
    ''' Given array x, and weights W, give median of X and min and max (1-sigma)
    '''
    # Normalize W
    W = W / np.sum(W)# Sort by x
    indices = np.argsort(x)
    x = x[indices]
    W = W[indices]
    cumsums = np.cumsum(W)
    # Find the 1-sigma lower limit
    index = (np.abs(cumsums-0.05)).argmin()
    xlow  = x[index]
    # Find the 1-sigma upper limit
    index = (np.abs(cumsums-0.95)).argmin()
    xhigh = x[index]
    # Find the median
    index  = (np.abs(cumsums-0.5)).argmin()
    xmedian= x[index]
    # Return values
    ##print("here",xmedian,xlow,xhigh)
    return xmedian, xmedian-xlow, xhigh-xmedian


##################################### 
# Compute the 90% confidence region for the $\mathcal{R}(z)$
##################################### 

str1=np.array(["gal","rpstoch_gal"])
str2=np.array(["min"])

nlim=300
ncases=2
RzAfterLensing = np.zeros((ncases,3, nlim))
RzBeforeLensing = np.zeros((ncases,3, nlim))

for ii in range(str1.size):
    for jj in range (str2.size):
        hh=jj+ (ii*str2.size)
        if(hh>4):
           break
        if(ii<1):
            Navgs1, kappa1, R0par1, W1 = np.loadtxt("all_o3_%s_rate_weights_mu%s.dat"%(str1[ii],str2[jj]),unpack=True)
        else:
            Navgs1, kappa1, gamma1, zp1, R0par1, W1 = np.loadtxt("all_o3_%s_rate_weights_mu%s.dat"%(str1[ii],str2[jj]),unpack=True)
        # Before lensing:
        zValues         = np.linspace(0,8,nlim)
        for kk in range(nlim):
            zi         = zValues[kk]
            if(ii<1):
                RzValueszi = RzPL(zi, kappa1,R0par1)
            else:
                RzValueszi = Rz(zi, kappa1,gamma1, zp1, R0par1)
            noWeight = np.ones(len(RzValueszi))
            RzMedian, RzLow, RzHigh = computeLimits(RzValueszi, noWeight)
            RzMin = RzMedian - RzLow
            RzMax = RzMedian + RzHigh
            # Save values
            RzBeforeLensing[hh,0,kk] = RzMedian
            RzBeforeLensing[hh,1,kk] = RzMin
            RzBeforeLensing[hh,2,kk] = RzMax
        # After lensing
            RzMedian, RzLow, RzHigh = computeLimits(RzValueszi, W1)
            RzMin = RzMedian - RzLow
            RzMax = RzMedian + RzHigh
            # Save values
            RzAfterLensing[hh,0,kk] = RzMedian
            RzAfterLensing[hh,1,kk] = RzMin
            RzAfterLensing[hh,2,kk] = RzMax

# Labels for different types of curves
clabels=np.array(["Gal","RzStoch + Lens"])
lsty=np.array(["dashdot","dashed"])


for ii in range(ncases):
    if(ii==0):
        x = zValues
        y1, y2 = RzBeforeLensing[ii,1,:], RzBeforeLensing[ii,2,:]
        plt.fill_between(x[x<=2.01], y1[x<=2.01], y2[x<=2.01 ], facecolor="grey",  interpolate=True,alpha=0.2, label="No lensing") ##facecolor="blue"
        plt.plot(x[x<=1.01],y1[x<=1.01], ls= "solid", color='black',alpha=0.7)
        plt.plot(x[ (x>=1.0) & (x<=2.01) ],y1[ (x>=1.0) & (x<=2.01) ], ls= ":", color='black',alpha=0.7)
        plt.plot(x[x<=1.01],y2[x<=1.01], ls= "solid", color='black',alpha=0.7)
        plt.plot(x[ (x>=1.0) & (x<=2.01) ],y2[ (x>=1.0) & (x<=2.01) ], ls= ":", color='black',alpha=0.7)
        
        y1, y2 = RzAfterLensing[ii,1,:], RzAfterLensing[ii,2,:]
        plt.fill_between(x[x<=2.01] , y1[x<=2.01 ], y2[x<=2.01], hatch="x" , facecolor="none", label="Lensing", interpolate=True,alpha=0.3) ##facecolor="red"


    if(ii>0):
        x = zValues

        y1, y2 = RzBeforeLensing[ii,1,:], RzBeforeLensing[ii,2,:]
        plt.plot(x,y1, ls= "solid", color='black',alpha=0.8, label="SGWB")
        plt.plot(x,y2, ls= "solid", color='black',alpha=0.8)
 
    x = zValues
    y1, y2 = RzAfterLensing[ii,1,:], RzAfterLensing[ii,2,:]
    if(ii==0):
        plt.plot(x,y1, ls=lsty[ii], color='black')
        plt.plot(x,y2, ls=lsty[ii], color='black')
 
    

plt.legend(loc="lower center", ncol=2)
plt.xlim([0,2.])
plt.ylim([1.0,5e3])
plt.xlabel(r"$z_{\rm m}$")
plt.ylabel(r"$\mathcal{R}_{\rm m}(z_{\rm m})$ [$\rm Gpc^{-3}$ $\rm yr^{-1}$]")
plt.yscale("log")
plt.grid()
plt.savefig(r"RateInf_cbc_woStochL_PL_zeq2.pdf", bbox_inches='tight')
